package no.ntnu.stud.frodesol.jexserver.eval;

import no.ntnu.stud.frodesol.jexserver.core.JexEnv;
import no.ntnu.stud.frodesol.jexserver.util.*;

import java.lang.Runtime;
import java.io.File;
import java.io.IOException;

public class Evaluator {
	
    static final int BUFFER_SIZE = 16384;
    
	private String m_exdir;
	private Exercise m_exercise;
	
	private int m_last_test_done = -1;
	private long m_test_started_at = 0;
	private long m_max_test_number = 0;
	
	public Evaluator (String exdir) {
		
		m_exdir = exdir;

	}
	
	public boolean run () {
		
		try {
			loadExercise ();
			m_max_test_number = m_exercise.getTree ().getLastNodeNumber ();
			initDirectories ();
			
	        ExerciseCompiler compiler = new ExerciseCompiler (m_exercise);
	        compiler.compileEverything ();
			
	        int start_from = 0;
			while (true) {
    	        try {
    	            m_last_test_done = start_from - 1;
    	            runJexTester (start_from);
    			    break;
    			}
    			catch (TimeoutException e) {
    			    System.out.println ("Detected timeout in test " + 
    			            e.getTestNumber ());
    			    start_from = e.getTestNumber () + 1;
    			    continue;
    			}
			}
			m_exercise.writeReport ();
			debugPrintExerciseTree ();
			debugPrintResult ();
		}
		catch (Exception e) {
			log ("Exception");
			log (e.getMessage ());
			log (e.getClass ().getName ());
			e.printStackTrace ();
			return false;
		}
		System.out.println ("run completed");
		return true;
	}
	
	
	
	private void initDirectories () throws IOException {
	    
	    new File (m_exercise.getSubmissionBinPath ()).mkdirs ();
	    File f = new File (m_exercise.getResultsDirectory ());
	    if (f.exists ()) {
	        Util.deleteFolder (f.getPath ());
	    }
	    f.mkdirs ();
	}
    
	private void runJexTester (int start_from) throws java.io.IOException, 
	        java.lang.InterruptedException, ExerciseException {
		
		
		System.out.println ("runJexTester");
		
		
		String cp = m_exercise.getTestBinPath () + File.pathSeparator +
		        JexEnv.getSystemClassPath () + File.pathSeparator +
		        m_exercise.getSubmissionBinPath () + File.pathSeparator +
		        m_exercise.getResourcePath ();
		
		String[] args = {
			JexEnv.getJavaCommand (),
			"-Djava.security.manager",
			"-Djava.security.policy=" + m_exercise.createPolicyFile (),
            "-Djava.io.tmpdir=" + m_exercise.createTempDir (),
			"-classpath",
			cp,
			"no.ntnu.stud.frodesol.jexserver.bin.JexTester",
			m_exercise.getPath (),
			Integer.toString (start_from)
		};
		
		for (int i = 0; i < args.length; i++) {
		    System.out.println (args[i]);
		}
		
		Runtime runtime = Runtime.getRuntime ();
		System.out.println ("Starting jextester from test " + start_from);
		Process p = runtime.exec (args);
		
		java.io.InputStream is = p.getInputStream ();
		java.io.InputStream es = p.getErrorStream ();
		
        byte[] buffer = new byte[BUFFER_SIZE];
        System.out.println (" -- jextester launched -- waiting for read --");
        
        restartTestTimer ();
        
        while (true) {
            try {
                int avail = is.available ();
                if (avail > 0) { 
                    int read = is.read (buffer, 0, Math.min (avail, BUFFER_SIZE));
                    if (read == -1) break;
                    
                    if (read > 0) {
                        System.out.write (buffer, 0, read);
                    }
                    
                }
                avail = es.available ();
                if (avail > 0) { 
                    int read = es.read (buffer, 0, Math.min (avail, BUFFER_SIZE));
                    if (read == -1) break;
                    
                    if (read > 0) {
                        System.out.write (buffer, 0, read);
                    }
                    
                }
            }
            catch (java.io.IOException e) {
                break;
            }
            
            if (getCurrentTestNo () > m_max_test_number) {
                // All tests are done
                break;
            }
            
            try {
                while (checkProgress ());
            }
            catch (TimeoutException e) {
                System.out.println ("TimeoutException, destroying " +
                        "jextester process");
                p.destroy ();
                throw e;
            }
            
            /*
            // Just sleep a bit here, to avoid a busy-loop using too much
            // CPU
            try {
                Thread.sleep (1);
            }
            catch (InterruptedException e) {}
            */
            
        }
        
        p.destroy();
        /*
        System.out.println ("Waiting for process to end");
        
        System.out.println ("(emptying error stream)");
        while (true) {
            try {
                int read = es.read (buffer, 0, 100);
                if (read <= 0) break;
            }
            catch (java.io.IOException e) {
                break;
            }
        }
        
        System.out.println ("(emptying input stream)");
        while (true) {
            try {
                int read = is.read (buffer, 0, 100);
                if (read <= 0) break;
            }
            catch (java.io.IOException e) {
                break;
            }
        }
        
        p.waitFor ();
        */
		System.out.println ("-- JEXTESTER EXITED --");
		
		// Just call checkProgress until it returns false -- when there is
		// no more test results to read
		while (checkProgress ());
		
	}
	
	private void restartTestTimer () {
	    m_test_started_at = System.currentTimeMillis ();
	}
	
	private boolean checkProgress () throws ExerciseException {
	    
	    int current_test = getCurrentTestNo ();
	    
	    File okfile = new File (m_exercise.getTestOkFilePath (current_test));
	    if (okfile.exists ()) {
	        m_exercise.readTestResult (current_test);
	        m_last_test_done = current_test;
	        restartTestTimer ();
	        return true;
	    }
	    else {
	        try {
	            TestNode node = m_exercise.getTree ().getNode (
	                    current_test);
	            if (System.currentTimeMillis () - m_test_started_at >
	                    node.getMaxRunningTimeMillis ()) {
	                // This test has exceeded its max running time
	                int seconds = node.getMaxRunningTimeMillis () / 1000;
	                node.addProblem ("Test terminated after " + seconds + 
	                        " seconds (infinite loop?)", 
	                        "The test was terminated after " +
	                        seconds + " seconds, most likely due to an " +
	                        "infinite loop. An infinite loop is a loop " +
	                        "(for/while) that never terminates.");
	                throw new TimeoutException (current_test);
	            }
	        }
	        catch (ArrayIndexOutOfBoundsException e) {
	            // This means that all tests are done, i.e. current_test
	            // refers to an invalid node 
	        }
	        return false;
	    }
	    
	}
	
	private int getCurrentTestNo () {
	    int next_test = m_last_test_done + 1;
	    return next_test;
	}
	
	private void loadExercise () throws ExerciseException {
		m_exercise = new Exercise (m_exdir);
	}
	
	private void printExerciseTreeNode (int level, TestNode node) {
	    String line = "";
	    if (node.getMaxPoints () > 0 && node.getAchievedPoints () == 0) {
            line += " -" + node.getMaxPoints ();
	    }
	    else {
	        line += "   ";
	    }
		for (int i = 0; i < level; i++) {
			line += "    ";
		}
		line += node.getName ();
        if (line.length () > 77) {
            line = line.substring (0, 75) + "...";
        }
        System.out.println (line);
		TestNode[] nodes = node.getChildren ();
		for (int i = 0; i < nodes.length; i++) {
			printExerciseTreeNode (level + 1, nodes[i]);
		}
	}
	
	private void debugPrintExerciseTree () {
		System.out.println ();
		printExerciseTreeNode (0, m_exercise.getTree ().getRootNode ());
	}
	
	private void debugPrintResult () {
	    int[] result = m_exercise.getTree ().calculateResult ();
	    System.out.println ("POINTS: "  + result[0] + " of " + result[1]);
	}
	
	public static void log (String message) {
		System.out.println (message);
	}
}
